import numpy as np


def GlobalAveragePool(output_count, X):
    spatial_shape = np.ndim(X) - 2
    Y = np.average(X, axis=tuple(range(spatial_shape, spatial_shape + 2)))
    for _ in range(spatial_shape):
        Y = np.expand_dims(Y, -1)
    return Y


def test():
    x = [[
           [[1, 2, 3],
            [1, 2, 3],
            [1, 2, 3]],
           [[1, 2, 3],
            [1, 2, 3],
            [1, 2, 3]],
          ]]

    x = np.array(x)
    print(x.shape)
    y = GlobalAveragePool(x)
    print(y)


if __name__ == "__main__":
    test()



import os
import argparse
import tempfile
import cProfile
from typing import Tuple, Any, Text, Sequence, Dict, Optional
import numpy

from onnx import numpy_helper
import onnx.checker
import onnx.onnx_cpp2py_export.checker as c_checker
from onnx import ModelProto, NodeProto, IR_VERSION

from .backend_rep import BackendRep
from .compiler import compile_from_model
from .opset import get_opset


class Backend(object):
    @classmethod
    def is_compatible(cls,
                      model,  # type: ModelProto
                      device='CPU',  # type: Text
                      **kwargs  # type: Any
                      ):  # type: (...) -> bool

        specs = []
        for i in range(len(model.opset_import)):
            opset_import = model.opset_import[i]
            specs.append({'domain': opset_import.domain, 'version': opset_import.version})

        opset = get_opset(specs)

        for i in range(len(model.graph.node)):
            if opset[model.graph.node[i].op_type] is None:
                # print('Not supported op_type:', model.graph.node[i].op_type)
                return False

        return True

    @classmethod
    def prepare(cls,
                model,  # type: ModelProto
                device='CPU',  # type: Text
                **kwargs  # type: Any
                ):  # type: (...) -> Optional[BackendRep]
        onnx.checker.check_model(model)

        if 'out' in kwargs and kwargs['out'] is not None:
            path = kwargs['out']
            os.makedirs(path, exist_ok=True)

            compile_from_model(model, path)
            return BackendRep(path)
        else:
            with tempfile.TemporaryDirectory() as path:
                compile_from_model(model, path)
                return BackendRep(path)

    @classmethod
    def run_model(cls,
                  model,  # type: ModelProto
                  inputs,  # type: Any
                  device='CPU',  # type: Text
                  **kwargs  # type: Any
                  ):  # type: (...) -> Tuple[Any, ...]
        backend = cls.prepare(model, device, **kwargs)
        return backend.run(inputs)

    @classmethod
    def run_node(cls,
                 node,  # type: NodeProto
                 inputs,  # type: Any
                 device='CPU',  # type: Text
                 outputs_info=None,  # type: Optional[Sequence[Tuple[numpy.dtype, Tuple[int, ...]]]]
                 **kwargs  # type: Dict[Text, Any]
                 ):  # type: (...) -> Optional[Tuple[Any, ...]]
        print('##### run_node')
        raise Exception('run_node is not implemented yet')
        '''Simple run one operator and return the results.
        Args:
            outputs_info: a list of tuples, which contains the element type and
            shape of each output. First element of the tuple is the dtype, and
            the second element is the shape. More use case can be found in
            https://github.com/onnx/onnx/blob/master/onnx/backend/test/runner/__init__.py
        '''
        # TODO Remove Optional from return type
        if 'opset_version' in kwargs:
            special_context = c_checker.CheckerContext()
            special_context.ir_version = IR_VERSION
            special_context.opset_imports = {'': kwargs['opset_version']}  # type: ignore
            onnx.checker.check_node(node, special_context)
        else:
            onnx.checker.check_node(node)

        specs = [{'domain': '', 'version': 15}]  # temporary code, please use special_context
        opset = get_opset(specs)

        output = None

        if node.op_type in opset:
            op = opset[node.op_type]
            output = op(*inputs)

        return (output.dtype, output.shape, output)

    @classmethod
    def supports_device(cls, device):  # type: (Text) -> bool
        return device in ['CPU', 'cpu']


def main(args):
    onnx_path = args.onnx[0]
    input_paths = args.pb
    output_dir = args.o

    model = onnx.load_model(onnx_path)
    inputs = []

    for input_path in input_paths:
        with open(input_path, 'rb') as f:
            tensor = onnx.TensorProto()
            tensor.ParseFromString(f.read())

            inputs.append(numpy_helper.to_array(tensor))

    backend = Backend.prepare(model, out=output_dir)
    outputs = backend.run(inputs)

    if type(outputs) == tuple:
        for output in outputs:
            print(output)
    else:
        print(outputs)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='ONNX Reference Backend')
    parser.add_argument('onnx', metavar='onnx', nargs=1, help='an input ONNX model file')
    parser.add_argument('pb', metavar='pb', nargs='*', help='tensor pb files')
    parser.add_argument('-o', metavar='output directory', type=str, nargs='?',
                        help='connx output directory(default is temporary directory)')
    parser.add_argument('-p', action='store_true', help='performance profiling')

    args = parser.parse_args()

    if args.p:
        print('Performance profiling...')
        cProfile.run('main(args)')
    else:
        main(args)



import numpy as np
from .util import Iterator


def _conv(Y, y_idx, X, x_iter, W, w_iter, batch, x_channel, w_channel, feature_map, dilations):
    feature_dim = len(X.shape) - 2
    feature_shape = X.shape[2:]
    kernel_shape = W.shape[2:]

    # Make dilation kernel
    new_kernel_shape = dilations * np.array(kernel_shape)
    if not np.all(dilations == 1):
        new_kernel_shape -= np.ones([feature_dim], dtype=np.int64)
    kernel = np.zeros(new_kernel_shape, dtype=W.dtype)

    slicers = []
    for i in range(feature_dim):
        slicers.append(slice(None, None, dilations[i]))

    kernel[tuple(slicers)] = W[feature_map, w_channel]

    while x_iter.next():
        x_idx = x_iter.index

        # Make padded patch of X[batch, channel]
        x_patch = np.zeros(kernel.shape)
        x_slicers, x_padded_slicers = [batch, x_channel], []

        # Make slicers for copy pactch of X[batch, channel] on to padded X
        for i in range(feature_dim):
            x_start = 0 if x_idx[i] < 0 else x_idx[i]
            x_end = min(feature_shape[i], x_idx[i] + kernel.shape[i])
            x_slicers.append(slice(x_start, x_end, None))

            x_padded_start = -x_idx[i] if x_idx[i] < 0 else 0
            x_padded_end = x_padded_start + (x_end - x_start)
            x_padded_slicers.append(slice(x_padded_start, x_padded_end, None))

        # Copy. Ex.           [0][0][0]    [0][0][0]
        #           [1][2] => [0][0][0] => [1][2][0]
        #           [3][4]    [0][0][0]    [3][4][0]
        x_patch[x_padded_slicers] = X[tuple(x_slicers)]

        # Convolute
        y = np.sum(x_patch.flatten() * kernel.flatten())
        Y[y_idx] += y
        y_idx += 1


# X: (N x C x H x W) N - batch, C - channel, H, W - feature 1, 2
# W: (M x C/group x kH x kW) M is number of feature Map
# B: (M)
# Y: (M x ( C x M ) x ...
def Conv(output_count, X, W, B, auto_pad, dilations, group, kernel_shape, pads, strides):
    # feature dimension
    feature_dim = len(X.shape) - 2

    # default attribute setting
    if len(dilations) == 0:
        dilations = np.ones([feature_dim], dtype=np.int64)
    else:
        dilations = np.array(dilations)

    kernel_shape = np.array(kernel_shape)

    if len(pads) == 0:
        pads = np.zeros([feature_dim * 2], dtype=np.int64)
    else:
        pads = np.array(pads)

    if len(strides) == 0:
        strides = np.ones([feature_dim], dtype=np.int64)
    else:
        strides = np.array(strides)

    # output_spatial_shape
    output_shape = np.zeros([feature_dim], dtype=np.int64)

    if auto_pad == 'SAME_LOWER' or auto_pad == 'SAME_UPPER':
        for i in range(feature_dim):
            output_shape[i] = np.ceil(X.shape[2 + i] / strides[i])
            pad = (output_shape[i] - 1) * strides[i] + ((kernel_shape[i] - 1) * dilations[i] + 1) - X.shape[2 + i]
            pads[i] = pads[i + feature_dim] = int(pad / 2)
            if pad % 2 == 1:
                if auto_pad == 'SAME_UPPER':
                    pads[i + feature_dim] += 1
                else:
                    pads[i] += 1
    else:
        for i in range(feature_dim):
            output_shape[i] = np.floor((X.shape[2 + i] + pads[i] + pads[i + feature_dim] -
                                       ((kernel_shape[i] - 1) * dilations[i] + 1)) / strides[i] + 1)

    # Conv
    Y = np.zeros([X.shape[0] * W.shape[0] * int(np.prod(output_shape))], dtype=X.dtype)

    y_idx = 0
    y_unit = np.prod(output_shape)

    x_iter = Iterator(-pads[0:feature_dim], -pads[0:feature_dim] + output_shape * strides, strides)
    w_iter = Iterator((0,) * len(kernel_shape), kernel_shape, (1,) * len(kernel_shape))

    for batch in range(X.shape[0]):
        for g in range(group):
            feature_group = int(W.shape[0] / group)
            for feature_map in range(g * feature_group, (g + 1) * feature_group):  # divide feature_maps into groups
                for channel in range(W.shape[1]):  # iterate all of channels of feature_map
                    _conv(Y, y_idx, X, x_iter, W, w_iter,
                          batch, g * W.shape[1] + channel, channel, feature_map, dilations)

                # Apply bias
                if B is not None:
                    Y[y_idx:y_idx + y_unit] += B[feature_map]

                # Next Y
                y_idx += y_unit

    y_shape = (X.shape[0], W.shape[0]) + tuple(output_shape)
    Y = Y.reshape(y_shape)

    return Y



import os
from typing import Tuple, Any
import numpy as np
import struct
from .opset import get_opset, get_argcount


class Graph:
    def __init__(self, backend, path, graph_id):
        self.backend = backend
        self.path = path
        self.graph_id = graph_id

        self.text = self.__read_text()
        self.initializer = []  # initializer 1 to n (initializer[0] contains value_info #1

        self.input = []
        self.output = []
        self.value_info = [None]

        self.__read_text()
        self.__read_initializer()
        # TODO: read_sparse_initializer

    def interprete(self, inputs):
        self._value_info(self.text[0])
        self._output(self.text[2])
        self._input(self.text[3])

        # Set initializer
        for idx in range(len(self.initializer)):
            self.value_info[idx + 1] = self.initializer[idx]

        # Set input
        for idx, id in zip(range(len(inputs)), self.input):
            self.value_info[id] = inputs[idx]

        # Execute node
        node_count = self._node(self.text[4])
        for i in range(5, 5 + node_count):
            self._exec(self.text[i])

        return [self.value_info[id] for id in self.output]

    def __read_text(self):
        file_path = os.path.join(self.path, '{}.text'.format(self.graph_id))
        with open(file_path, 'r') as f:
            return f.readlines()

    def __read_initializer(self):
        tokens = list(self.text[1].split(' '))
        tokens.pop(0)
        count = int(tokens.pop(0))

        for i in range(1, count + 1):
            file_path = os.path.join(self.path, '{}_{}.data'.format(self.graph_id, i))
            tensor = self.__read_tensor(file_path)
            self.initializer.append(tensor)

    def __read_tensor(self, path):
        with open(path, 'rb') as f:
            dtype, ndim = struct.unpack('=II', f.read(8))
            shape = []
            for i in range(ndim):
                shape.append(struct.unpack('=I', f.read(4))[0])

            np_dtype = (None, np.float32, np.uint8, np.int8, np.uint16, np.int16,
                        np.int32, np.int64, str, bool, np.float16, np.uint32,
                        np.uint64, np.csingle, np.cdouble, None)[dtype]

            if np_dtype is None:
                raise Exception('Tensor data type {} is not supported yet'.format(np_dtype))

            buf = f.read()
            return np.frombuffer(buf, dtype=np_dtype).reshape(shape)

    def _value_info(self, line):
        tokens = list(line.split(' '))
        tokens.pop(0)

        # Empty values with null (0th index)
        self.value_info = [None] * (int(tokens.pop(0)) + 1)

    def _output(self, line):
        tokens = list(line.split(' '))
        tokens.pop(0)

        count = int(tokens.pop(0))

        for i in range(count):
            self.output.append(int(tokens.pop(0)))

    def _input(self, line):
        tokens = list(line.split(' '))
        tokens.pop(0)

        count = int(tokens.pop(0))

        for i in range(count):
            self.input.append(int(tokens.pop(0)))

    def _node(self, line):
        tokens = list(line.split(' '))
        tokens.pop(0)

        return int(tokens.pop(0))

    def _exec(self, line):
        tokens = list(line.split(' '))

        op_type = tokens.pop(0)
        output_count = int(tokens.pop(0))
        input_count = int(tokens.pop(0))
        attribute_count = int(tokens.pop(0))

        # parse output IDs
        output = []
        for i in range(output_count):
            output.append(int(tokens.pop(0)))

        # parse input IDs
        input = []
        for i in range(input_count):
            input.append(int(tokens.pop(0)))

        # parse attribute values
        attribute = []
        for i in range(attribute_count):
            _ = tokens.pop(0)  # drop name length
            _ = tokens.pop(0)  # drop attribute name
            attr_type = int(tokens.pop(0))

            if attr_type == 1:  # FLOAT
                value = float(tokens.pop(0))
            elif attr_type == 2:  # INT
                value = int(tokens.pop(0))
            elif attr_type == 3:  # STRING
                count = int(tokens.pop(0))
                value = ''
                while len(value) < count:
                    if len(value) != 0:
                        value += ' '
                    value += tokens.pop(0)
            elif attr_type == 6:  # FLOATS
                count = int(tokens.pop(0))
                value = []
                for i in range(count):
                    value.append(float(tokens.pop(0)))
            elif attr_type == 7:  # INTS
                count = int(tokens.pop(0))
                value = []
                for i in range(count):
                    value.append(int(tokens.pop(0)))
            elif attr_type == 8:  # STRINGS
                count = int(tokens.pop(0))
                value = []
                for i in range(count):
                    count2 = int(tokens.pop(0))
                    value2 = ''

                    while len(value2) < count2:
                        if len(value2) != 0:
                            value2 += ' '
                        value2 += tokens.pop(0)

                    value.append(value2)
            else:
                raise Exception('AttributeType: {} is not supported yet'.format(attr_type))

            attribute.append(value)

        # Get Operator
        op = self.backend.opset[op_type]
        if op is None:
            raise Exception('op_type {} is not supported yet'.format(op_type))

        argcount = self.backend.argcount[op_type]

        # Make argument for the operator
        # check minimum input count
        if len(input) < argcount[0]:
            raise Exception('op_type {} must have at least {} args but {}'.format(op_type, argcount[0], len(input)))

        args = [output_count]
        if argcount[1] != -1:  # argcount[1] == -1 means maximum argument count will be unlimited
            for i in range(argcount[1]):
                if i < len(input):
                    args.append(self.value_info[input[i]])
                else:
                    args.append(None)
        else:
            args.extend([self.value_info[id] for id in input])

        args.extend(attribute)

        # Execute the operator
        result = op(*args)

        # Set output
        if type(result) is tuple:
            for i, id in zip(range(len(result)), output):
                self.value_info[id] = result[i]
        else:
            self.value_info[output[0]] = result

        return True


class BackendRep(object):
    def __init__(self, path):
        self.opset = None
        self.graph_count = 0

        self.text = {}
        self.data = {}

        # parse connx
        self.__parse_connx(os.path.join(path, 'model.connx'))

        # read text
        for graph_id in range(self.graph_count):
            self.text[graph_id] = Graph(self, path, graph_id)

    def __parse_connx(self, path):
        lines = self.__read_text(path)

        # check connx version
        tokens = lines[0].split(' ')
        if tokens.pop(0) != 'connx' or int(tokens.pop(0)) != 3:
            raise Exception('not supported connx version: {}'.format(lines[0].trim()))

        # parse opset_import
        tokens = lines[1].split(' ')
        specs = []
        tokens.pop(0)
        for i in range(int(tokens.pop(0))):
            domain_len = int(tokens.pop(0))
            domain = tokens.pop(0)
            while len(domain) < domain_len:
                domain += ' ' + tokens.pop(0)

            version = int(tokens.pop(0))

            specs.append({'domain': domain, 'version': version})

        self.opset = get_opset(specs)
        self.argcount = get_argcount(specs)

        # parse graph
        tokens = lines[2].split(' ')
        tokens.pop(0)
        self.graph_count = int(tokens.pop(0))

    def __read_text(self, path):
        with open(path, 'r') as f:
            return f.readlines()

    def run(self, inputs, **kwargs):  # type: (Any, **Any) -> Tuple[Any, ...]
        graph = self.text[0]
        return graph.interprete(inputs)


import sys
import os
import itertools
import onnx
from onnx import numpy_helper
from .opset import get_attrset
import struct

#For ref optional
class _GlobalRefPrint(object) :
    _ref_print = 0
class _RefGetterSetter(_GlobalRefPrint) :
    def set_ref_print(self, ref_print):
        _GlobalRefPrint._ref_print = ref_print
    

class ConnxObject:
    def __init__(self, proto=None, parent=None):
        self.proto = proto
        self.parent = parent

    def _tab(self, out, depth):
        for i in range(depth):
            out.write('\t')

    def get_root(self):
        node = self

        while node.parent is not None:
            node = node.parent

        return node

    def dump(self, depth):
        pass

    def compile(self):
        pass


class ConnxModelProto(ConnxObject):
    def __init__(self, proto, arg_ref):
        super().__init__(proto)

        self.next_graph_id = 0

        #set ref value
        refClass = _RefGetterSetter()
        refClass.set_ref_print(arg_ref)
        
        # Parse opset_import
        specs = []
        for i in range(len(proto.opset_import)):
            opset_import = proto.opset_import[i]
            specs.append({'domain': opset_import.domain, 'version': opset_import.version})

        self.attrset = get_attrset(specs)

        # parse
        self.graph = ConnxGraphProto(proto.graph, self)

    def alloc_graph_id(self):
        id = self.next_graph_id
        self.next_graph_id += 1

        return id

    def dump(self, depth=0):
        out = sys.stdout

        self._tab(out, depth)
        out.write('ModelProto\n')

        self._tab(out, depth + 1)
        out.write('opset_import ')
        out.write(str(len(self.proto.opset_import)))
        out.write(' ')

        for i in range(len(self.proto.opset_import)):
            opset_import = self.proto.opset_import[i]
            domain = opset_import.domain
            version = opset_import.version

            out.write(str(len(domain)))
            out.write(' ')
            out.write(domain)
            out.write(' ')
            out.write(str(version))
            out.write('\n')

        self.graph.dump(depth + 1)

    def compile(self, path):
        os.makedirs(path, exist_ok=True)

        # Write graph
        self.graph.compile(path)

        with open(os.path.join(path, 'model.connx'), 'w') as out:
            # Write connx version
            out.write('connx 1\n')

            # Write opset_import
            out.write('opset_import ')
            out.write(str(len(self.proto.opset_import)))
            out.write(' ')

            for i in range(len(self.proto.opset_import)):
                opset_import = self.proto.opset_import[i]
                domain = opset_import.domain
                version = opset_import.version

                out.write(str(len(domain)))
                out.write(' ')
                out.write(domain)
                out.write(' ')
                out.write(str(version))
                out.write('\n')

            # Write number of graphs
            out.write('graph ')
            out.write(str(self.next_graph_id))
            out.write('\n')


class ConnxGraphProto(ConnxObject):
    def __init__(self, proto, parent):
        super().__init__(proto, parent)

        self.id = self.get_root().alloc_graph_id()

        self.initializer = [ConnxTensorProto(proto, self) for proto in proto.initializer]
        self.sparse_initializer = [ConnxSparseTensorProto(proto, self) for proto in proto.sparse_initializer]
        self.input = []
        self.output = []
        self.value_info = [ConnxValueInfoProto(proto, self) for proto in proto.value_info]

        # Get ref_print optional value
        ref_print = _GlobalRefPrint._ref_print

        # Make value_info reflects to initializer
        for initializer in self.initializer:
            value_info = self.get_value_info(initializer.proto.name)

            if value_info is None:
                value_info_proto = onnx.ValueInfoProto()
                value_info_proto.name = initializer.proto.name
                value_info_proto.type.tensor_type.elem_type = initializer.proto.data_type
                for i in range(len(initializer.proto.dims)):
                    dim = onnx.TensorShapeProto.Dimension()
                    dim.dim_value = initializer.proto.dims[i]
                    value_info_proto.type.tensor_type.shape.dim.append(dim)

                value_info = ConnxValueInfoProto(value_info_proto, self)
                value_info.initializer = initializer
                self.value_info.append(value_info)

        # Make value_info reflects to sparse_initializer
        for sparse_initializer in self.sparse_initializer:
            value_info = self.get_value_info(sparse_initializer.proto.name)

            if value_info is None:
                value_info_proto = onnx.ValueInfoProto()
                value_info_proto.name = sparse_initializer.proto.name
                value_info_proto.type.tensor_type.elem_type = sparse_initializer.proto.data_type

                for i in range(len(sparse_initializer.proto.dims)):
                    dim = onnx.TensorShapeProto.Dimension()
                    dim.dim_value = sparse_initializer.proto.dims[i]
                    value_info_proto.type.tensor_type.shape.dim.append(dim)

                value_info = ConnxValueInfoProto(value_info_proto, self)
                value_info.sparse_initializer = sparse_initializer
                self.value_info.append(value_info)

        # Make value_info reflects to input
        for input in proto.input:
            if input.name == '':
                continue

            value_info = self.get_value_info(input.name)

            if value_info is None:
                value_info = ConnxValueInfoProto(input, self)
                self.value_info.append(value_info)
            self.input.append(value_info)

        # Make value_info reflects to output
        for output in proto.output:
            if output.name == '':
                continue

            value_info = self.get_value_info(output.name)

            if value_info is None:
                value_info = ConnxValueInfoProto(output, self)
                self.value_info.append(value_info)

            self.output.append(value_info)

        # Make value_info reflects to node
        for node in proto.node:
            for name in itertools.chain(node.input, node.output):
                value_info = self.get_value_info(name)
                if name == '':
                    continue
                if value_info is None:
                    value_info = ConnxValueInfoProto(None, self, name=name)
                    self.value_info.append(value_info)

        # Assign ID to value_info, which has intializer and sparse_initializer first
        id = 1
        for value_info in self.value_info:
            if value_info.initializer is not None:
                value_info.id = id
                value_info.initializer.id = id
                id += 1
            elif value_info.sparse_initializer is not None:
                value_info.id = id
                value_info.sparse_initializer.id = id
                id += 1

        # Assign ID to value_info, which doesn't have intializer and sparse_initializer next
        for value_info in self.value_info:
            if value_info.initializer is None and value_info.sparse_initializer is None:
                value_info.id = id
                id += 1

        self.node = [ConnxNodeProto(proto, self) for proto in proto.node ]
        temp_initalizer_input_count = 0
        # Internal operators
        for initializer in self.initializer:
            idx = 0
            temp_initalizer_input_count += 1
            ref_count = self.get_ref_count(initializer.proto.name)
            if ref_print == 1 :
                if ref_count > 1:
                    print('create proto_{}'.format(temp_initalizer_input_count))
                    node = self.make_ref_count(initializer.proto.name, initializer.id, ref_count)
                    self.node.insert(idx, node)
                    idx += 1
            else : 
                if ref_count == 1 :
                    idx += 1
            
        for sparse_initializer in self.sparse_initializer:
            idx = 0
            ref_count = self.get_ref_count(sparse_initializer.proto.name)
            if ref_count > 1:
                node = self.make_ref_count(sparse_initializer.proto.name, sparse_initializer.id, ref_count)
                self.node.insert(idx, node)
                idx += 1
        idx = 0
        length = len(self.node)
        while idx < length:
            node = self.node[idx]
            for name in node.proto.output:
                ref_count = self.get_ref_count(name, idx + 1)
                if ref_print == 1 :
                    if ref_count > 1:
                        value_info = self.get_value_info(name)
                        node = self.make_ref_count(name, value_info.id, ref_count)
                        self.node.insert(idx + 1, node)
                        idx += 1
                        length += 1
                else :
                    if ref_count > 1 :
                        idx += 1
            idx += 1

    def get_value_info(self, name):
        for value_info in self.value_info:
            if value_info.proto is not None and value_info.proto.name == name or value_info.name == name:
                return value_info

        return None

    def dump(self, depth):
        out = sys.stdout

        self._tab(out, depth)
        out.write('GraphProto\n')

        self._tab(out, depth + 1)
        out.write('id ')
        out.write(str(self.id))
        out.write('\n')

        self._tab(out, depth + 1)
        out.write('name ')
        out.write(self.proto.name)
        out.write('\n')

        self._tab(out, depth + 1)
        out.write('initializer\n')

        for initializer in self.initializer:
            initializer.dump(depth + 2)

        self._tab(out, depth + 1)
        out.write('sparse_initializer\n')

        for sparse_initializer in self.sparse_initializer:
            sparse_initializer.dump(depth + 2)

        self._tab(out, depth + 1)
        out.write('input\n')

        for input in self.input:
            input.dump(depth + 2)

        self._tab(out, depth + 1)
        out.write('output\n')

        for output in self.output:
            output.dump(depth + 2)

        self._tab(out, depth + 1)
        out.write('value_info\n')

        for value_info in self.value_info:
            value_info.dump(depth + 2)

        self._tab(out, depth + 1)
        out.write('node\n')

        for node in self.node:
            node.dump(depth + 2)

    def get_ref_count(self, name, idx=0):
        count = 0

        for i in range(idx, len(self.node)):
            if name in self.node[i].proto.input:
                count += 1

        if name in self.proto.output:
            count += 1

        return count

    def make_ref_count(self, name, id, ref_count):
        attr = onnx.AttributeProto()
        attr.name = 'ref_count'
        attr.type = onnx.AttributeProto.AttributeType.INT
        attr.i = ref_count

        proto = onnx.NodeProto()
        proto.name = '_ref'
        proto.op_type = '_ref'
        proto.input.append(name)
        proto.attribute.append(attr)

        return ConnxNodeProto(proto, self)

    def compile(self, path):
        def write(tensor, array, out):
            out.write(struct.pack('=I', tensor.proto.data_type))  # dtype
            out.write(struct.pack('=I', len(tensor.proto.dims)))  # ndim

            for i in range(len(tensor.proto.dims)):  # dims
                out.write(struct.pack('=I', tensor.proto.dims[i]))

            buf = array.tobytes()
            out.write(buf)

        for initializer in self.initializer:
            name = '{}_{}.data'.format(str(self.id), str(initializer.id))

            with open(os.path.join(path, name), 'wb') as out:
                array = numpy_helper.to_array(initializer.proto)
                write(initializer, array, out)

        for sparse_initializer in self.sparse_initializer:
            name = '{}_{}.data'.format(str(self.id), str(sparse_initializer.id))

            # TODO: Compress array
            with open(os.path.join(path, name), 'wb') as out:
                array = numpy_helper.to_array(sparse_initializer.proto)
                write(initializer, array, out)

        # graph name
        with open(os.path.join(path, str(self.id) + '.text'), 'w') as out:
            # variable count
            out.write('value_info ')
            out.write(str(len(self.value_info)))
            out.write('\n')

            # initializer count
            out.write('initializer ')
            out.write(str(len(self.initializer) + len(self.sparse_initializer)))
            out.write('\n')

            # output count id id, ...
            out.write('output ')
            out.write(str(len(self.output)))

            for i in range(len(self.output)):
                out.write(' ')
                out.write(str(self.output[i].id))

            out.write('\n')

            # input count id id, ...
            out.write('input ')
            out.write(str(len(self.input)))

            for i in range(len(self.input)):
                out.write(' ')
                out.write(str(self.input[i].id))

            out.write('\n')

            # node count
            out.write('node ')
            out.write(str(len(self.node)))
            out.write('\n')

            for node in self.node:
                node.compile(out)


class ConnxTensorProto(ConnxObject):
    def __init__(self, proto, parent):
        super().__init__(proto, parent)

        self.id = 0

    def dump(self, depth):
        out = sys.stdout

        self._tab(out, depth)
        out.write('TensorProto\n')

        self._tab(out, depth + 1)
        out.write('name ')
        out.write(self.proto.name)
        out.write('\n')

        self._tab(out, depth + 1)
        out.write('data_type ')
        out.write(ConnxTensorProto.DataType(self.proto.data_type))
        out.write('\n')

        self._tab(out, depth + 1)
        out.write('dims')
        for i in range(len(self.proto.dims)):
            out.write(' ')
            out.write(str(self.proto.dims[i]))
        out.write('\n')

    def DataType(data_type):
        if data_type == 1:
            return 'FLOAT'
        elif data_type == 2:
            return 'UINT8'
        elif data_type == 3:
            return 'INT8'
        elif data_type == 4:
            return 'UINT16'
        elif data_type == 5:
            return 'INT16'
        elif data_type == 6:
            return 'INT32'
        elif data_type == 7:
            return 'INT64'
        elif data_type == 8:
            return 'STRING'
        elif data_type == 9:
            return 'BOOL'
        elif data_type == 10:
            return 'FLOAT16'
        elif data_type == 11:
            return 'DOUBLE'
        elif data_type == 12:
            return 'UINT32'
        elif data_type == 13:
            return 'UINT64'
        elif data_type == 14:
            return 'COMPLEX64'
        elif data_type == 15:
            return 'COMPLEX128'
        elif data_type == 16:
            return 'BFLOAT16'
        else:
            return 'UNDEFINED'


class ConnxSparseTensorProto(ConnxObject):
    def __init__(self, proto, parent):
        super().__init__(proto, parent)

        self.id = 0

    def dump(self, depth):
        out = sys.stdout

        self._tab(out, depth)
        out.write('SparseTensorProto\n')


class ConnxValueInfoProto(ConnxObject):
    def __init__(self, proto, parent, name=None):
        super().__init__(proto, parent)

        self.id = 0
        self.type = ConnxTypeProto(proto.type, self) if proto is not None else None
        self.name = name
        self.initializer = None
        self.sparse_initializer = None

    def dump(self, depth):
        out = sys.stdout

        self._tab(out, depth)
        out.write('ValueInfoProto\n')

        self._tab(out, depth + 1)
        out.write('name ')
        if self.proto is not None:
            out.write(self.proto.name)
        else:
            out.write('<generated>')
        out.write('\n')

        if self.type is not None:
            self._tab(out, depth + 1)
            out.write('type\n')
            self.type.dump(depth + 2)


class ConnxTypeProto(ConnxObject):
    def __init__(self, proto, parent):
        super().__init__(proto, parent)

    def dump(self, depth):
        out = sys.stdout

        self._tab(out, depth)
        out.write('TypeProto\n')

        if self.proto.tensor_type is not None:
            self._tab(out, depth + 1)
            out.write('tensor_type\n')
            tensor_type = self.proto.tensor_type

            self._tab(out, depth + 2)
            out.write('elem_type ')
            out.write(ConnxTensorProto.DataType(tensor_type.elem_type))
            out.write('\n')

            self._tab(out, depth + 2)
            out.write('shape')

            for i in range(len(tensor_type.shape.dim)):
                dim = tensor_type.shape.dim[i]

                out.write(' ')
                if dim.dim_param != '':
                    out.write(dim.dim_param)
                else:
                    out.write(str(dim.dim_value))

            out.write('\n')
        elif self.proto.sequence_type is not None:
            out.write('WARNING: sequence type is not supported')
        elif self.proto.map_type is not None:
            out.write('WARNING: map type is not supported')


class ConnxAttributeProto(ConnxObject):
    def __init__(self, proto, parent):
        super().__init__(proto, parent)

    def dump(self, depth):
        out = sys.stdout

        self._tab(out, depth)
        out.write('AttributeProto\n')

        self._tab(out, depth + 1)
        out.write('name ')
        out.write(self.proto.name)
        out.write('\n')

        self._tab(out, depth + 1)
        out.write('type ')
        out.write(ConnxAttributeProto.AttributeType(self.proto.type))
        out.write('\n')

        self._tab(out, depth + 1)
        if self.proto.type in [1, 2, 3, 6, 7, 8]:
            out.write('value ')
            out.write(str(self.value()))
            out.write('\n')
        else:
            out.write('value\n')
            self.value().dump(depth + 2)

    def compile(self, out):
        if self.proto.type == 0:
            pass  # UNDEFINED means null attribute
        else :
            out.write(str(len(self.proto.name)))
            out.write(' ')
            out.write(self.proto.name)
            out.write(' ')
            out.write(str(self.proto.type))
            out.write(' ')

            if self.proto.type in [1, 2]:
                out.write(str(self.value()))
            elif self.proto.type == 3:
                s = self.value().decode('utf-8')
                out.write(str(len(s)))
                out.write(' ')
                out.write(s)
            elif self.proto.type in [6, 7]:
                value = self.value()
                out.write(str(len(value)))

                for i in range(len(value)):
                    out.write(' ')
                    out.write(str(value[i]))
            elif self.proto.type == 8:
                value = self.value()
                out.write(str(len(value)))

                for i in range(len(value)):
                    s = value[i].decode('utf-8')
                    out.write(' ')
                    out.write(str(len(str(s))))
                    out.write(' ')
                    out.write(str(s))
            else:
                raise 'Not implemented yet: AttributeType ' + str(self.proto.type)

    def value(self):
        type = self.proto.type
        if type == 1:
            return self.proto.f
        elif type == 2:
            return self.proto.i
        elif type == 3:
            return self.proto.s
        elif type == 4:
            # TODO: Proto를 동적으로 만들면 안되고, value_info에 초기에 초기화 한 후 id를 할당 받아야 함. 그리고 실제 활용할 때는 id로 접근해야 함
            return ConnxTensorProto(self.proto.t, None)
        elif type == 5:
            # TODO: graph의 경우 id 또는 name으로 접근해야 함
            # value에선 당연히 숫자 또는 문자를 넘겨야 함
            return ConnxGraphProto(self.proto.g, None)
        elif type == 11:
            return ConnxSparseTensorProto(self.proto.sparse_tensor, None)
        elif type == 6:
            return self.proto.floats
        elif type == 7:
            return self.proto.ints
        elif type == 8:
            return self.proto.strings
        elif type == 9:
            return [ConnxTensorProto(self.proto.tensors[i]) for i in range(len(self.proto.tensors))]
        elif type == 10:
            return [ConnxGraphProto(self.proto.graphs[i]) for i in range(len(self.proto.graphs))]
        elif type == 12:
            return [ConnxSparseTensorProto(self.proto.sparse_tensors[i]) for i in range(len(self.proto.sparse_tensors))]
        else:
            return None

    def AttributeType(type):
        if type == 1:
            return 'FLOAT'
        elif type == 2:
            return 'INT'
        elif type == 3:
            return 'STRING'
        elif type == 4:
            return 'TENSOR'
        elif type == 5:
            return 'GRAPH'
        elif type == 11:
            return 'SPARSE_TENSOR'
        elif type == 6:
            return 'FLOATS'
        elif type == 7:
            return 'INTS'
        elif type == 8:
            return 'STRINGS'
        elif type == 9:
            return 'TENSORS'
        elif type == 10:
            return 'GRAPHS'
        elif type == 12:
            return 'SPARSE_TENSORS'
        else:
            return 'UNDEFINED'


class ConnxNodeProto(ConnxObject):
    def __init__(self, proto, parent):
        super().__init__(proto, parent)
        def get_attribute_proto(name):
            for attribute_proto in proto.attribute:
                if attribute_proto.name == name:
                    return attribute_proto

            return None

        self.attribute = []

        root = self.get_root()
        attrset = root.attrset[proto.op_type]

        for default_attr in attrset:
            original_attr = get_attribute_proto(default_attr.name)
            if original_attr is not None:
                self.attribute.append(ConnxAttributeProto(original_attr, self))
            else:
                self.attribute.append(ConnxAttributeProto(default_attr, self))

    def dump(self, depth):
        out = sys.stdout

        self._tab(out, depth)
        out.write('NodeProto\n')

        self._tab(out, depth + 1)
        out.write('name ')
        out.write(self.proto.name)
        out.write('\n')

        self._tab(out, depth + 1)
        out.write('op_type ')
        out.write(self.proto.op_type)
        out.write('\n')

        self._tab(out, depth + 1)
        out.write('input')
        for i in range(len(self.proto.input)):
            out.write(' ')
            out.write(self.proto.input[i])
        out.write('\n')

        self._tab(out, depth + 1)
        out.write('output')
        for i in range(len(self.proto.output)):
            out.write(' ')
            out.write(self.proto.output[i])
        out.write('\n')

        self._tab(out, depth + 1)
        out.write('attribute\n')
        for attribute in self.attribute:
            attribute.dump(depth + 2)
        out.write('\n')

    def compile(self, out):
        out.write(self.proto.op_type)
        out.write(' ')

        out.write(str(len(self.proto.output)))
        out.write(' ')

        out.write(str(len(self.proto.input)))
        out.write(' ')
        temp = len(self.attribute)
        for i in range(0, len(self.attribute)):
            if self.attribute[i].value() is None:
                temp = temp - 1
        out.write(str(temp))

        for i in range(len(self.proto.output)):
            value_info = self.parent.get_value_info(self.proto.output[i])
            out.write(' ')
            if value_info is not None:
                out.write(str(value_info.id))
            else:
                out.write('0')

        for i in range(len(self.proto.input)):
            value_info = self.parent.get_value_info(self.proto.input[i])
            out.write(' ')
            if value_info is not None:
                out.write(str(value_info.id))
            else:
                out.write('0')

        for attribute in self.attribute:
            if attribute.value() is None:
                continue
            out.write(' ')
            attribute.compile(out)

        out.write('\n')



import numpy as np
from typing import Any, List, Callable, Optional, Text


def cartesian(arrays, out=None):
    # type: (List[np.ndarray], np.ndarray) -> np.ndarray
    """
    From https://stackoverflow.com/a/1235363
    Generate a cartesian product of input arrays.
    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.
    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.
    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])
    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = n // arrays[0].size
    out[:, 0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m, 1:])
        for j in range(1, arrays[0].size):
            out[j * m:(j + 1) * m, 1:] = out[0:m, 1:]
    return out


def interpolate_1d_with_x(data,                                             # type: np.ndarray
                          scale_factor,                                     # type: float
                          x,                                                # type: float
                          get_coeffs,                                       # type: Callable[[float], np.ndarray]
                          roi=None,                                         # type: np.ndarray
                          extrapolation_value=0.0,                          # type: float
                          coordinate_transformation_mode='half_pixel',      # type: Text
                          exclude_outside=False,                            # type: bool
                          ):                                                # type: (...) -> np.ndarray
    def get_neighbor_idxes(x, n, limit):  # type: (float, int, int) -> np.ndarray
        """
        Return the n nearest indexes to x among [0, limit), prefer the indexes smaller than x.
        As a result, the ratio must be in (0, 1]
        Examples:
        get_neighbor_idxes(4, 2, 10) == [3, 4]
        get_neighbor_idxes(4, 3, 10) == [3, 4, 5]
        get_neighbor_idxes(4.4, 3, 10) == [3, 4, 5]
        get_neighbor_idxes(4.5, 3, 10) == [3, 4, 5]
        get_neighbor_idxes(4.6, 3, 10) == [4, 5, 6]
        get_neighbor_idxes(4.4, 1, 10) == [4]
        get_neighbor_idxes(4.6, 1, 10) == [5]
        :param x:
        :param n: the number of the wanted indexes
        :param limit: the maximum value of index
        :return: An np.array containing n nearest indexes in ascending order
        """
        idxes = sorted(range(limit), key=lambda idx: (abs(x - idx), idx))[:n]
        idxes = sorted(idxes)
        return np.array(idxes)

    def get_neighbor(x, n, data):  # type: (float, int, np.ndarray) -> np.ndarray
        """
        Pad `data` in 'edge' mode, and get n nearest elements in the padded array and their indexes in the original
        array
        :param x: center index (in the unpadded coordinate system) of the found nearest elements.
        :param n: the number of neighbors.
        :param data: the array
        :return: A tuple containing the indexes of neighbor elements (the index can be smaller than 0 or higher than
        len(data)) and the value of these elements
        """
        pad_width = np.ceil(n / 2).astype(np.int)
        padded = np.pad(data, pad_width, mode='edge')
        x += pad_width

        idxes = get_neighbor_idxes(x, n, len(padded))
        ret = padded[idxes]
        return idxes - pad_width, ret

    input_width = len(data)
    output_width = scale_factor * input_width
    if coordinate_transformation_mode == 'align_corners':
        if output_width == 1:
            x_ori = 0.
        else:
            x_ori = x * (input_width - 1) / (output_width - 1)
    elif coordinate_transformation_mode == 'asymmetric':
        x_ori = x / scale_factor
    elif coordinate_transformation_mode == 'tf_crop_and_resize':
        if output_width == 1:
            x_ori = (roi[1] - roi[0]) * (input_width - 1) / 2
        else:
            x_ori = x * (roi[1] - roi[0]) * \
                (input_width - 1) / (output_width - 1)
        x_ori += (roi[0] * (input_width - 1))
        # Return extrapolation_value directly as what TF CropAndResize does
        if x_ori < 0 or x_ori > input_width - 1:
            return extrapolation_value
    elif coordinate_transformation_mode == 'tf_half_pixel_for_nn':
        x_ori = (x + 0.5) / scale_factor
    elif coordinate_transformation_mode == 'pytorch_half_pixel':
        if output_width == 1:
            x_ori = -0.5
        else:
            x_ori = (x + 0.5) / scale_factor - 0.5
    else:  # coordinate_transformation_mode == 'half_pixel'
        x_ori = (x + 0.5) / scale_factor - 0.5
    x_ori_int = np.floor(x_ori).astype(np.int).item()

    # ratio must be in (0, 1] since we prefer the pixel on the left of `x_ori`
    if x_ori.is_integer():
        ratio = 1
    else:
        ratio = x_ori - x_ori_int

    coeffs = get_coeffs(ratio)
    n = len(coeffs)

    idxes, points = get_neighbor(x_ori, n, data)

    if exclude_outside:
        for i, idx in enumerate(idxes):
            if idx < 0 or idx >= input_width:
                coeffs[i] = 0
        coeffs /= sum(coeffs)

    return np.dot(coeffs, points).item()


def interpolate_nd_with_x(data,                      # type: np.ndarray
                          n,                         # type: int
                          scale_factors,             # type: List[float]
                          x,                         # type: List[float]
                          get_coeffs,                # type: Callable[[float], np.ndarray]
                          roi=None,                  # type: np.ndarray
                          **kwargs                   # type: Any
                          ):                         # type: (...) -> np.ndarray
    if n == 1:
        return interpolate_1d_with_x(data, scale_factors[0], x[0], get_coeffs, roi=roi,
                                     **kwargs)
    return interpolate_1d_with_x(
        [interpolate_nd_with_x(data[i], n - 1, scale_factors[1:], x[1:], get_coeffs,
                               roi=None if roi is None else np.concatenate(
                                   [roi[1:n], roi[n + 1:]]),
                               **kwargs)
         for i in range(data.shape[0])], scale_factors[0], x[0], get_coeffs,
        roi=None if roi is None else [roi[0], roi[n]], **kwargs)


def interpolate_nd(data,                      # type: np.ndarray
                   get_coeffs,                # type: Callable[[float], np.ndarray]
                   output_size=None,          # type: Optional[List[int]]
                   scale_factors=None,        # type: Optional[List[float]]
                   roi=None,                  # type: np.ndarray
                   **kwargs                   # type: Any
                   ):                         # type: (...) -> np.ndarray
    def get_all_coords(data):   # type: (np.ndarray) -> np.ndarray
        return cartesian([list(range(data.shape[i])) for i in range(len(data.shape))])

    assert output_size is not None or scale_factors is not None
    if output_size is not None:
        scale_factors = np.array(output_size) / np.array(data.shape)
    else:
        output_size = (scale_factors * np.array(data.shape)).astype(np.int)
    assert scale_factors is not None

    ret = np.zeros(output_size)
    for x in get_all_coords(ret):
        ret[tuple(x)] = interpolate_nd_with_x(data, len(data.shape), scale_factors, x, get_coeffs, roi=roi,
                                              **kwargs)
    return ret


def cubic_coeffs(ratio, A=-0.75):   # type: (float, float) -> np.ndarray
    coeffs = [((A * (ratio + 1) - 5 * A) * (ratio + 1) + 8 * A) * (ratio + 1) - 4 * A,
              ((A + 2) * ratio - (A + 3)) * ratio * ratio + 1,
              ((A + 2) * (1 - ratio) - (A + 3)) * (1 - ratio) * (1 - ratio) + 1,
              ((A * ((1 - ratio) + 1) - 5 * A) * ((1 - ratio) + 1) + 8 * A) * ((1 - ratio) + 1) - 4 * A]

    return np.array(coeffs)


def linear_coeffs(ratio):           # type: (float) -> np.ndarray
    return np.array([1 - ratio, ratio])


def nearest_coeffs(ratio, mode='round_prefer_floor'):          # type: (float, Text) -> np.ndarray
    mode = mode.rstrip()
    if type(ratio) == int or ratio.is_integer():
        return np.array([0, 1])
    elif mode == 'round_prefer_floor':
        return np.array([ratio <= 0.5, ratio > 0.5])
    elif mode == 'round_prefer_ceil':
        return np.array([ratio < 0.5, ratio >= 0.5])
    elif mode == 'floor':
        return np.array([1, 0])
    elif mode == 'ceil':
        return np.array([0, 1])


def Resize(output_count, X, roi, scales, sizes, coordinate_transformation_mode, cubic_coeff_a, exclude_outside,
           extrapolation_value, mode, nearest_mode):
    if mode == "nearest":
        return interpolate_nd(X,
                              lambda x: nearest_coeffs(x, mode=nearest_mode),
                              output_size=sizes,
                              roi=roi,
                              scale_factors=scales,
                              coordinate_transformation_mode=coordinate_transformation_mode,
                              extrapolation_value=extrapolation_value,
                              exclude_outside=exclude_outside).astype(np.float32)
    elif mode == "linear":
        return interpolate_nd(X,
                              linear_coeffs,
                              output_size=sizes,
                              roi=roi,
                              scale_factors=scales,
                              coordinate_transformation_mode=coordinate_transformation_mode,
                              extrapolation_value=extrapolation_value,
                              exclude_outside=exclude_outside).astype(np.float32)
    elif mode == "cubic":
        return interpolate_nd(X,
                              lambda x: cubic_coeffs(x, A=cubic_coeff_a),
                              output_size=sizes,
                              roi=roi,
                              scale_factors=scales,
                              coordinate_transformation_mode=coordinate_transformation_mode,
                              extrapolation_value=extrapolation_value,
                              exclude_outside=exclude_outside).astype(np.float32)
    else:
        raise NotImplementedError



import numpy as np


def BatchNormalization(output_count, X, scale, B, input_mean, input_var, epsilon, momentum, training_mode):
    dims_x = len(X.shape)
    dim_ones = (1,) * (dims_x - 2)
    scale = scale.reshape(-1, *dim_ones)
    B = B.reshape(-1, *dim_ones)
    input_mean = input_mean.reshape(-1, *dim_ones)
    input_var = input_var.reshape(-1, *dim_ones)

    Y = (X - input_mean) / np.sqrt(input_var + epsilon) * scale + B

    return Y


def test():
    x = np.random.randn(2, 3, 4, 5).astype(np.float32)
    s = np.random.randn(3).astype(np.float32)
    bias = np.random.randn(3).astype(np.float32)
    mean = np.random.randn(3).astype(np.float32)
    var = np.random.rand(3).astype(np.float32)

    y1 = BatchNormalization(x, s, bias, mean, var, 1e-05, 0.9, 1)
    print(y1)
    y2, output_mean, output_var = BatchNormalization(x, s, bias, mean, var, 1e-05, 0.9, 1)
    print(y2, output_mean, output_var)


if __name__ == "__main__":
    test()



import sys
import os
import argparse
import numpy as np
import onnx
from onnx import numpy_helper
import struct
from .proto import ConnxModelProto


def load_model(path, ref_print) -> ConnxModelProto:
    proto = onnx.load_model(path)

    return ConnxModelProto(proto, ref_print)


def compile_from_model(model_proto, path) -> int:
    connx = ConnxModelProto(model_proto)
    connx.compile(path)

    return 0


def load_tensor(path) -> onnx.TensorProto:
    tensor = onnx.TensorProto()
    with open(path, 'rb') as f:
        tensor.ParseFromString(f.read())

    return tensor


def compile(*_args: str) -> int:
    parser = argparse.ArgumentParser(description='ONNX-CONNX Command Line Interface')
    parser.add_argument('onnx', metavar='onnx', nargs='+', help='an input ONNX model file or tensor pb file')
    parser.add_argument('-d', action='store_true', help='dump human readable onnx metadata to standard output')
    parser.add_argument('-o', metavar='output directory', type=str, default='out', nargs='?',
                        help='output directory(default is out)')
    parser.add_argument('--ref', metavar='ref print boolean', type=int, default=0, help="default(no print) --> --ref 0 || print ref --> --ref 1")
    # parser.add_argument('-p', metavar='profile', type=str, nargs='?', help='specify configuration file')
    # parser.add_argument('-c', metavar='comment', type=str, nargs='?', choices=['true', 'false', 'True', 'False'],
    #                     help='output comments(true or false)')

    # parse args
    ref_print_arg = 0
    if len(_args) > 0:
        args = parser.parse_args(_args)
    else:
        args = parser.parse_args()

    for path in args.onnx:
        if path.endswith('.onnx'):
            if args.ref :
                ref_print_arg = args.ref
            model = load_model(path,ref_print_arg)

            if args.d:
                model.dump()
            else:
                model.compile(args.o)
        elif path.endswith('.pb'):
            tensor = load_tensor(path)

            if args.d:
                array = numpy_helper.to_array(tensor)

                np.set_printoptions(suppress=True, threshold=sys.maxsize, linewidth=160)
                print(array)
            else:
                name = os.path.basename(path).strip('.pb') + '.data'

                with open(os.path.join(args.o, name), 'wb') as out:
                    out.write(struct.pack('=I', tensor.data_type))
                    out.write(struct.pack('=I', len(tensor.dims)))
                    for i in range(len(tensor.dims)):
                        out.write(struct.pack('=I', tensor.dims[i]))

                    array = numpy_helper.to_array(tensor)
                    out.write(array.tobytes())

    return 0

def convert(modelpath, outpath = 'out', isdump = None, isref = None) -> int:
    if isref == 1 :
        ref_print_arg = isref
    else :
        isref == 0 or isref == None
        ref_print_arg = 0
        
    model = load_model(modelpath,ref_print_arg)
    print(model)
    if isdump == 1 :
        model.dump()
    else :
        model.compile(outpath)


import unittest
import onnx.backend.test
from .backend import Backend


pytest_plugins = 'onnx.backend.test.report',
backend_test = onnx.backend.test.runner.Runner(Backend, __name__)

# Exclude training operators
backend_test.exclude(r'test_adagrad_*')
backend_test.exclude(r'test_adam_*')
backend_test.exclude(r'test_gradient_*')
backend_test.exclude(r'test_momentum_*')
backend_test.exclude(r'nesterov_momentum_*')
backend_test.exclude(r'test_batchnorm_epsilon_training_mode')
backend_test.exclude(r'test_batchnorm_example_training_mode')
backend_test.exclude(r'test_identity_sequence_cpu')

# Exclude deprecated operators
backend_test.exclude(r'test_scatter_*')
backend_test.exclude(r'test_upsample_*')

# Exclude legacy onnx api
backend_test.exclude(r'test_operator_clip')  # We support above Clip-6

# import all test cases at global scope to make them visible to python.unittest
globals().update(backend_test.enable_report().test_cases)

if __name__ == '__main__':
    unittest.main()



import onnx
import numpy as np

def _null(name):
    attr = onnx.AttributeProto()
    attr.name = name
    attr.type = onnx.AttributeProto.AttributeType.UNDEFINED
    return attr

def _float(name, value = None):
    if value is None :
        return _null(name)
    attr = onnx.AttributeProto()
    attr.name = name
    attr.type = onnx.AttributeProto.AttributeType.FLOAT
    attr.f = value
    return attr


def _int(name, value = None):
    if value is None :
        return _null(name)
    else :
        attr = onnx.AttributeProto()
        attr.name = name
        attr.type = onnx.AttributeProto.AttributeType.INT
        attr.i = value
    return attr


def _string(name, value):
    attr = onnx.AttributeProto()
    attr.name = name
    attr.type = onnx.AttributeProto.AttributeType.STRING
    attr.s = str.encode(value, 'utf-8')
    return attr


def _floats(name, value):
    attr = onnx.AttributeProto()
    attr.name = name
    attr.type = onnx.AttributeProto.AttributeType.FLOATS
    attr.floats.extend(value)
    return attr


def _ints(name, value):
    attr = onnx.AttributeProto()
    attr.name = name
    attr.type = onnx.AttributeProto.AttributeType.INTS
    attr.ints.extend(value)
    return attr


def _strings(name, value):
    attr = onnx.AttributeProto()
    attr.name = name
    attr.type = onnx.AttributeProto.AttributeType.STRINGS
    attr.strings.extend([str.encode(s, 'utf-8') for s in value])
    return attr


class Iterator:
    # start: array of start position
    # stop: array of stop position
    # step: array of step position
    # All the length must be same
    def __init__(self, start, stop, step):
        self.ndim = len(start)
        self.start = [*start]
        self.stop = [*stop]
        self.step = [*step]

        assert self.ndim == len(self.stop)
        assert self.ndim == len(self.step)

        self.index = np.array([*self.start])
        self.index[-1] -= self.step[-1]

    def next(self):
        # Go next step
        for i in range(self.ndim - 1, -1, -1):
            self.index[i] += self.step[i]
            if self.index[i] < self.stop[i]:
                return True
            else:
                self.index[i] = self.start[i]

        # Return to just before start
        self.index = np.array([*self.start])
        self.index[-1] -= self.step[-1]

        return False

    # Get offset from shape
    def offset(self, shape):
        return _index_to_offset(self.index, shape)


def _index_to_offset(index, shape):
    offset = 0
    unit = 1
    for i in range(len(shape) - 1, -1, -1):
        offset += unit * index[i]
        unit *= shape[i]

    return offset


def _offset_to_index(offset, shape):
    index = []
    for i in range(1, len(shape)):
        unit = np.prod(np.array(shape[i:]))
        idx, offset = divmod(offset, unit)
        index.append(idx)
    index.append(offset)
    return tuple(index)


__all__ = ['_int', '_float', '_string', '_ints', '_floats',
           '_strings', 'Iterator', '_index_to_offset', '_offset_to_index']



from numpy import equal
from .opset import version as default_version
from .opset import opset as default_opset
from .opset import argcount as default_argcount
from .opset import attrset as default_attrset


_versions = {
    '': default_version
}

_opsets = {
    '': default_opset
}

_argcounts = {
    '': default_argcount
}

_attrsets = {

    '': default_attrset
}

# specs: [ { domain: str, version: int } ]
def _check(specs):
    for spec in specs:
        domain = spec['domain']
        version = spec['version']
        if domain not in _versions:
            _versions[domain] = default_version
            _attrsets[domain] = default_attrset

        if _versions[domain] < version:
            raise Exception('The opset version is lower than the expected one: opset domain: {}, version: {}, '
                            'expected: {}'.format(domain, _versions[domain], version))


def get_opset(specs):
    _check(specs)

    opset = {}

    for spec in specs:
        domain = spec['domain']
        _ = spec['version']

        opsets = _opsets[domain]
        for key, value in opsets.items():
            opset[key] = value

    return opset


def get_argcount(specs):
    _check(specs)

    argcount = {}

    for spec in specs:
        domain = spec['domain']
        _ = spec['version']

        argcounts = _argcounts[domain]

        for key, value in argcounts.items():
            argcount[key] = value

    return argcount


def get_attrset(specs):
    _check(specs)

    attrset = {}

    for spec in specs:
        domain = spec['domain']
        _ = spec['version']
        attrsets = _attrsets[domain]
    for key, value in attrsets.items():
        attrset[key] = value

    return attrset


__all__ = [get_opset, get_argcount, get_attrset]



import os
import subprocess
import setuptools


# # version
# if os.path.isdir('.git'):
#     cwd = os.path.dirname(os.path.realpath(__file__))

#     __version_raw = subprocess.run(['git', 'describe', '--tags', '--long'],
#                                    stdout=subprocess.PIPE, cwd=cwd).stdout.decode('utf-8')[1:]
#     __versions = __version_raw.split('-')
#     __version__ = ''
#     for i in range(0,len(__versions)) :
#         __version__ =__version__ + __versions[i]
#         if i != len(__versions) :
#             __version__ = __version__ + " "

#     with open('VERSION', 'w') as f:
#         f.write(__version__)
# else:
#     with open('VERSION', 'r') as f:
#         __version__ = f.read()

# description
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

# setup
setuptools.setup(name='onnx-connx-neowine',
                 version= 0.0,
                 author='Semih Kim',
                 author_email='semih.kim@gmail.com',
                 description='ONNX to CONNX converting tool',
                 long_description=open('README.md').read(), 
                 long_description_content_type='text/markdown',
                 url='https://github.com/semihlab/onnx-connx',
                 project_urls={
                     'Bug Tracker': 'https://github.com/semihlab/onnx-connx/issues',
                 },
                 classifiers=[
                     'Environment :: Console',
                     'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
                     'Topic :: Software Development :: Build Tools',
                 ],
                 package_dir={'': '.'},
                 packages=setuptools.find_packages(where='.'),
                 tests_require=['pytest'],
                 entry_points={
                     'console_scripts': [
                         'onnx2connx=onnx_connx.cli:main',
                     ],
                 })

                # install_requires=[
                #      'onnx',
                #      'python3',
                #      'pip'
                #  ],


import numpy as np
from .util import _int, _float, _string, _ints, _null, _floats, _strings
from .MaxPool import MaxPool
from .BatchNormalization import BatchNormalization
from .GlobalAveragePool import GlobalAveragePool
from .Conv import Conv
from .Cast import Cast
from .Resize import Resize


# Internal operators
def _ref(output_count, tensor):
    pass


# Most of the implementations are fllowd by ONNX reference implementation
def Abs(ouput_count, X):
    return np.abs(X)


def Acos(output_count, X):
    return np.arccos(X)


def Acosh(ouput_count, input):
    return np.arccosh(input)


def Add(ouput_count, A, B):
    return A + B


def And(output_count, A, B):
    return np.logical_and(A, B)


def ArgMax(output_count, data, axis, keepdims, select_last_index):
    if select_last_index == 1:
        data = np.flip(data, axis)

    result = np.argmax(data, axis=axis)
    if keepdims == 1:
        result = np.expand_dims(result, axis)

    if select_last_index == 1:
        result = data.shape[axis] - result - 1

    return result.astype(np.int64)


def ArgMin(output_count, data, axis, keepdims, select_last_index):
    if select_last_index == 1:
        data = np.flip(data, axis)

    result = np.argmin(data, axis=axis)
    if keepdims == 1:
        result = np.expand_dims(result, axis)

    if select_last_index == 1:
        result = data.shape[axis] - result - 1

    return result.astype(np.int64)


def Asin(output_count, input):
    return np.arcsin(input)


def MatMul(output_count, A, B):

    return np.matmul(A, B)


def Relu(output_count, X):
    return np.clip(X, 0, np.inf)


def Reshape(ouput_count, data, shape, allowzero):
    new_shape = np.copy(shape)

    if allowzero == 0:
        zeros_index = np.where(shape == 0)
        new_shape[zeros_index] = np.array(data.shape)[zeros_index]

    reshaped = np.reshape(data, new_shape)

    return reshaped


def Concat(*inputs):
    r"""
    inputs type constraints
        tensor(uint8), tensor(uint16), tensor(uint32), tensor(uint64),
        tensor(int8), tensor(int16), tensor(int32), tensor(int64),
        tensor(bfloat16), tensor(float16), tensor(float), tensor(double),
        tensor(string), tensor(bool), tensor(complex64), tensor(complex128)
    """
    axis = inputs[-1]
    inputs = inputs[1:-1]
    concat_result = np.concatenate(inputs, axis)

    return concat_result


def Exp(output_count, input):
    r"""
    input type constraints
        tensor(float16), tensor(float), tensor(double), tensor(bfloat16)
    """
    return np.exp(input)


def Flatten(output_count, input, axis):
    r"""
        :param input: A Tensor of rank >= axis.
        :param axis: Indicate up to which input dimensions (exclusive) should be flattened to the outer dimension of
                     the output.

        input type constraints
            tensor(uint8), tensor(uint16), tensor(uint32), tensor(uint64), tensor(int8), tensor(int16),
            tensor(int32), tensor(int64), tensor(bfloat16), tensor(float16), tensor(float), tensor(double),
            tensor(string), tensor(bool), tensor(complex64), tensor(complex128)
    """
    if axis < 0:
        axis = len(input.shape) + axis

    front = 1
    end = 1
    if axis != 0:
        for i in range(0, len(input.shape)):
            if i >= axis:
                end *= input.shape[i]
            else:
                front *= input.shape[i]
        return input.reshape(front, end)
    else:
        flatten = 1
        for x in input.shape:
            flatten *= x
        return np.array([input.reshape(flatten)])


def Gather(output_count, data, indices, axis):
    r"""
    :param data: Tensor of rank r >= 1.
    :param indices: Tensor of int32/int64 indices of any rank q.
                    All index values are expected to be
                    within bounds [-s, s-1] along axis of size s.
                    It is an error if any of the index values are out of bounds.
    """
    return np.take(data, indices, axis)


def Identity(output_count, input):
    r"""
    :param input: Input tensor.

    input type constraints
        tensor(uint8), tensor(uint16), tensor(uint32), tensor(uint64), tensor(int8), tensor(int16),
        tensor(int32), tensor(int64), tensor(bfloat16), tensor(float16), tensor(float), tensor(double),
        tensor(string), tensor(bool), tensor(complex64), tensor(complex128), seq(tensor(uint8)),
        seq(tensor(uint16)), seq(tensor(uint32)), seq(tensor(uint64)), seq(tensor(int8)), seq(tensor(int16)),
        seq(tensor(int32)), seq(tensor(int64)), seq(tensor(float16)), seq(tensor(float)), seq(tensor(double)),
        seq(tensor(string)), seq(tensor(bool)), seq(tensor(complex64)), seq(tensor(complex128))
    """
    if type(input) == list:
        return [np.copy(i) for i in input]
    else:
        return np.copy(input)


def LeakyRelu(ouput_count, X, alpha):
    return np.clip(X, 0, np.inf) + np.clip(X, -np.inf, 0) * alpha


def Log(output_count, input):
    return np.log(input)


def Mul(output_count, A, B):
    return A * B


def Shape(output_count, data):
    return np.array(data.shape, dtype=np.int64)


def Sigmoid(output_count, X):
    return 1.0 / (1.0 + np.exp(np.negative(X)))


def Slice(output_count, data, starts, ends, axes, steps):
    # TODO : Naive implementation.
    expr = [":"] * len(data.shape)
    if axes is None:
        axes = np.array([i for i in range(len(data.shape))])

    for i in range(axes.shape[0]):
        if steps is None:
            expr[axes[i]] = f"{starts[i]}:{ends[i]}"
        else:
            expr[axes[i]] = f"{starts[i]}:{ends[i]}:{steps[i]}"

    slice_str = (",").join(expr)
    eval_str = f"data[{slice_str}]"

    return eval(eval_str)


def Split(output_count, input, split, axis):
    outputs = []
    start = 0

    if split is None:
        return tuple(np.array_split(input, output_count, axis))

    for i in range(split.shape[0]):
        expr = [":"] * len(input.shape)
        expr[axis] = f"{start}:{start + split[i]}"
        slice_str = (",").join(expr)
        eval_str = f"input[{slice_str}]"
        outputs.append(eval(eval_str))
        start = split[i]

    return tuple(outputs)


def Tanh(output_count, X):
    return np.tanh(X)


def Transpose(output_count, data, perm):
    if not perm:
        return np.transpose(data)
    return np.transpose(data, perm)


def Clip(output_count, X, min_val, max_val):
    if min_val is None:
        if np.issubdtype(X.dtype, np.integer):
            min_val = np.iinfo(X.dtype).min
        elif np.issubdtype(X.dtype, np.floating):
            min_val = -np.inf

    if max_val is None:
        if np.issubdtype(X.dtype, np.integer):
            max_val = np.iinfo(X.dtype).max
        elif np.issubdtype(X.dtype, np.floating):
            max_val = np.inf

    return np.clip(X, min_val, max_val)

version = 18

opset = {
    'QLinearConcat' : None,
    '_ref': _ref,
    'Abs': Abs,
    'Acos': Acos,
    'Acosh': Acosh,
    'Add': Add,
    'And': And,
    'ArgMax': ArgMax,
    'ArgMin': ArgMin,
    'Asin': Asin,
    'Asinh': None,
    'Atan': None,
    'Atanh': None,
    'AveragePool': None,
    'BatchNormalization': BatchNormalization,
    'BitShift': None,
    'Cast': Cast,
    'Ceil': None,
    'Celu': None,
    'Clip': Clip,
    'Compress': None,
    'Concat': Concat,
    'ConcatFromSequence': None,
    'Constant': None,
    'ConstantOfShape': None,
    'Conv': Conv,
    'ConvInteger': None,
    'ConvTranspose': None,
    'Cos': None,
    'Cosh': None,
    'CumSum': None,
    'DepthToSpace': None,
    'DequantizeLinear': None,
    'Det': None,
    'Div': None,
    'Dropout': None,
    'Einsum': None,
    'Elu': None,
    'Equal': None,
    'Erf': None,
    'Exp': Exp,
    'Expand': None,
    'EyeLike': None,
    'Flatten': Flatten,
    'Floor': None,
    'GRU': None,
    'Gather': Gather,
    'GatherElements': None,
    'GatherND': None,
    'Gemm': None,
    'GlobalAveragePool': GlobalAveragePool,
    'GlobalLpPool': None,
    'GlobalMaxPool': None,
    'Greater': None,
    'HardSigmoid': None,
    'Hardmax': None,
    'Identity': Identity,
    'If': None,
    'InstanceNormalization': None,
    'IsInf': None,
    'IsNaN': None,
    'LRN': None,
    'LSTM': None,
    'LeakyRelu': LeakyRelu,
    'Less': None,
    'Log': Log,
    'Loop': None,
    'LpNormalization': None,
    'LpPool': None,
    'MatMul': MatMul,
    'MatMulInteger': None,
    'Max': None,
    'MaxPool': MaxPool,
    'MaxRoiPool': None,
    'MaxUnpool': None,
    'Mean': None,
    'Min': None,
    'Mod': None,
    'Mul': Mul,
    'Multinomial': None,
    'Neg': None,
    'NonMaxSuppression': None,
    'NonZero': None,
    'Not': None,
    'OneHot': None,
    'Or': None,
    'PRelu': None,
    'Pad': None,
    'Pow': None,
    'QLinearConv': None,
    'QLinearMatMul': None,
    'QuantizeLinear': None,
    'RNN': None,
    'RandomNormal': None,
    'RandomNormalLike': None,
    'RandomUniform': None,
    'RandomUniformLike': None,
    'Reciprocal': None,
    'ReduceL1': None,
    'ReduceL2': None,
    'ReduceLogSum': None,
    'ReduceLogSumExp': None,
    'ReduceMax': None,
    'ReduceMean': None,
    'ReduceMin': None,
    'ReduceProd': None,
    'ReduceSum': None,
    'ReduceSumSquare': None,
    'Relu': Relu,
    'Reshape': Reshape,
    'Resize': Resize,
    'ReverseSequence': None,
    'RoiAlign': None,
    'Round': None,
    'Scan': None,
    'Scatter (deprecated)': None,
    'ScatterElements': None,
    'ScatterND': None,
    'Selu': None,
    'SequenceAt': None,
    'SequenceConstruct': None,
    'SequenceEmpty': None,
    'SequenceErase': None,
    'SequenceInsert': None,
    'SequenceLength': None,
    'Shape': Shape,
    'Shrink': None,
    'Sigmoid': Sigmoid,
    'Sign': None,
    'Sin': None,
    'Sinh': None,
    'Size': None,
    'Slice': Slice,
    'Softplus': None,
    'Softsign': None,
    'SpaceToDepth': None,
    'Split': Split,
    'SplitToSequence': None,
    'Sqrt': None,
    'Squeeze': None,
    'StringNormalizer': None,
    'Sub': None,
    'Sum': None,
    'Tan': None,
    'Tanh': Tanh,
    'TfIdfVectorizer': None,
    'ThresholdedRelu': None,
    'Tile': None,
    'TopK': None,
    'Transpose': Transpose,
    'Trilu': None,
    'Unique': None,
    'Unsqueeze': None,
    'Upsample (deprecated)': None,
    'Where': None,
    'Xor': None,
    'Function': None,
    'DynamicQuantizeLinear': None,
    'GreaterOrEqual': None,
    'HardSwish': None,
    'LessOrEqual': None,
    'LogSoftmax': None,
    'MeanVarianceNormalization': None,
    'NegativeLogLikelihoodLoss': None,
    'Range': None,
    'Softmax': None,
    'SoftmaxCrossEntropyLoss': None,
    'QLinearAdd' : None,
    'QLinearGlobalAveragePool' : None,
}

#  minimum input count, maximum input count]
# -1 means unlimited
argcount = {
    '_ref': [1, 1],
    'Abs': [1, 1],
    'Acos': [1, 1],
    'Acosh': [1, 1],
    'Add': [2, 2],
    'And': [2, 2],
    'ArgMax': [1, 1],
    'ArgMin': [1, 1],
    'Asin': [1, 1],
    'Asinh': [1, 1],
    'Atan': [1, 1],
    'Atanh': [1, 1],
    'AveragePool': [1, 1],
    'BatchNormalization': [5, 5],
    'BitShift': [2, 2],
    'Cast': [1, 1],
    'Ceil': [1, 1],
    'Celu': [1, 1],
    'Clip': [1, 3],
    'Compress': [2, 2],
    'Concat': [1, -1],
    'ConcatFromSequence': None,
    'Constant': None,
    'ConstantOfShape': None,
    'Conv': [2, 3],
    'ConvInteger': [2,4],
    'ConvTranspose': None,
    'Cos': None,
    'Cosh': None,
    'CumSum': None,
    'DepthToSpace': None,
    'DequantizeLinear': [2,3],
    'Det': None,
    'Div': None,
    'Dropout': None,
    'Einsum': None,
    'Elu': None,
    'Equal': None,
    'Erf': None,
    'Exp': [1, 1],
    'Expand': None,
    'EyeLike': None,
    'Flatten': [1, 1],
    'Floor': None,
    'GRU': None,
    'Gather': [2, 2],
    'GatherElements': None,
    'GatherND': None,
    'Gemm': [2,3],
    'GlobalAveragePool': [1, 1],
    'GlobalLpPool': None,
    'GlobalMaxPool': None,
    'Greater': None,
    'HardSigmoid': None,
    'Hardmax': None,
    'Identity': [1, 1],
    'If': None,
    'InstanceNormalization': None,
    'IsInf': None,
    'IsNaN': None,
    'LRN': None,
    'LSTM': [3,8],
    'LeakyRelu': [1, 1],
    'Less': None,
    'Log': [1, 1],
    'Loop': None,
    'LpNormalization': None,
    'LpPool': None,
    'MatMul': [2, 2],
    'MatMulInteger': [2,4],
    'Max': None,
    'MaxPool': [1, 1],
    'MaxRoiPool': None,
    'MaxUnpool': None,
    'Mean': None,
    'Min': None,
    'Mod': None,
    'Mul': [2, 2],
    'Multinomial': None,
    'Neg': None,
    'NonMaxSuppression': None,
    'NonZero': None,
    'Not': None,
    'OneHot': None,
    'Or': None,
    'PRelu': None,
    'Pad': None,
    'Pow': None,
    'QLinearConv': [8,9],
    'QLinearMatMul': [8,8],
    'QuantizeLinear': [2,3],
    'RNN': [3,6],
    'RandomNormal': None,
    'RandomNormalLike': None,
    'RandomUniform': None,
    'RandomUniformLike': None,
    'Reciprocal': None,
    'ReduceL1': None,
    'ReduceL2': None,
    'ReduceLogSum': None,
    'ReduceLogSumExp': None,
    'ReduceMax': None,
    'ReduceMean': None,
    'ReduceMin': None,
    'ReduceProd': None,
    'ReduceSum': None,
    'ReduceSumSquare': None,
    'Relu': [1, 1],
    'Reshape': [2, 2],
    'Resize': [1, 4],
    'ReverseSequence': None,
    'RoiAlign': None,
    'Round': None,
    'Scan': None,
    'Scatter (deprecated)': None,
    'ScatterElements': None,
    'ScatterND': None,
    'Selu': None,
    'SequenceAt': None,
    'SequenceConstruct': None,
    'SequenceEmpty': None,
    'SequenceErase': None,
    'SequenceInsert': None,
    'SequenceLength': None,
    'Shape': [1, 1],
    'Shrink': None,
    'Sigmoid': [1, 1],
    'Sign': None,
    'Sin': None,
    'Sinh': None,
    'Size': None,
    'Slice': [3, 5],
    'Softplus': None,
    'Softsign': None,
    'SpaceToDepth': None,
    'Split': [1, 2],
    'SplitToSequence': None,
    'Sqrt': None,
    'Squeeze': None,
    'StringNormalizer': None,
    'Sub': None,
    'Sum': None,
    'Tan': None,
    'Tanh': [1, 1],
    'TfIdfVectorizer': None,
    'ThresholdedRelu': None,
    'Tile': None,
    'TopK': None,
    'Transpose': [1, 1],
    'Trilu': None,
    'Unique': None,
    'Unsqueeze': [2,2],
    'Upsample (deprecated)': None,
    'Where': None,
    'Xor': None,
    'Function': None,
    'DynamicQuantizeLinear': None,
    'GreaterOrEqual': None,
    'HardSwish': None,
    'LessOrEqual': None,
    'LogSoftmax': None,
    'MeanVarianceNormalization': None,
    'NegativeLogLikelihoodLoss': None,
    'Range': None,
    'Softmax': None,
    'SoftmaxCrossEntropyLoss': None,
    'QLinearAdd' : [7,8],
    'QLinearGlobalAveragePool' : [5,5],
    'QLinearConcat' : [3,-1],
}

attrset = {
    '_ref': [_int('ref_count', 0)],
    'Abs': [],
    'Acos': [],
    'Acosh': [],
    'Add': [],
    'And': [],
    'ArgMax': [_int('axis', 0), _int('keepdims', 1), _int('select_last_index', 0)],
    'ArgMin': [_int('axis', 0), _int('keepdims', 1), _int('select_last_index', 0)],
    'Asin': [],
    'Asinh': [],
    'Atan': [],
    'Atanh': [],
    'AveragePool': [_string('auto_pad', 'NOTSET'), _int('ceil_mode', 0), _int('count_include_pad', 0),
                    _ints('kernel_shape', []), _ints('pads', []), _ints('strides', [])],
    'BatchNormalization': [_float('epsilon', 1e-05), _float('momentum', 0.9), _int('training_mode', 0)],
    'BitShift': [_string('direction', '')],
    'Cast': [_int('to', 0)],
    'Ceil': [],
    'Celu': [_float('alpha', 1.0)],
    'Clip': [],
    'Compress': [_int('axis', 0)],
    'Concat': [_int('axis', 0)],
    'ConcatFromSequence': [_int('axis', 0), _int('new_axis', 0)],
    'Constant': [],
    'ConstantOfShape': [],
    'Conv': [_string('auto_pad', 'NOTSET'), _ints('dilations', []), _int('group', 1),
             _ints('kernel_shape', []), _ints('pads', []), _ints('strides', [])],
    'ConvInteger': [_string('auto_pad', 'NOTSET'), _ints('dilations', []), _int('group', 1),
             _ints('kernel_shape', []), _ints('pads', []), _ints('strides', [])],
    'ConvTranspose': [],
    'Cos': [],
    'Cosh': [],
    'CumSum': [],
    'DepthToSpace': [],
    'DequantizeLinear': [_int('axis', 1)],
    'Det': [],
    'Div': [],
    'Dropout': [],
    'Einsum': [],
    'Elu': [],
    'Equal': [],
    'Erf': [],
    'Exp': [],
    'Expand': [],
    'EyeLike': [],
    'Flatten': [_int('axis', 1)],
    'Floor': [],
    'GRU': [],
    'Gather': [_int('axis', 0)],
    'GatherElements': [],
    'GatherND': [],
    'Gemm': [_float('alpha', 1.0), _float('beta', 1.0), _int('transA', 0), _int('transB', 0)],
    'GlobalAveragePool': [],
    'GlobalLpPool': [],
    'GlobalMaxPool': [],
    'Greater': [],
    'HardSigmoid': [_float('alpha',0.2), _float('beta',0.5)],
    'Hardmax': [],
    'Identity': [],
    'If': [],
    'InstanceNormalization': [],
    'IsInf': [],
    'IsNaN': [],
    'LRN': [],
    'LSTM': [_floats('activation_alpha', []), _floats('activation_beta', []), _strings('activation', []), _float('clip'), _string('direction', 'forward'), _int('hidden_size'), _int('input_forget', 0), _int('Layout',0)],
    'LeakyRelu': [_float('alpha', 0.01)],
    'Less': [],
    'Log': [],
    'Loop': [],
    'LpNormalization': [],
    'LpPool': [],
    'MatMul': [],
    'MatMulInteger': [],
    'Max': [],
    'MaxPool': [_string('auto_pad', 'NOTSET'), _int('ceil_mode', 0), _ints('dilations', []),
                _ints('kernel_shape', []), _ints('pads', []), _int('storage_order', 0), _ints('strides', [])],
    'MaxRoiPool': [],
    'MaxUnpool': [],
    'Mean': [],
    'Min': [],
    'Mod': [],
    'Mul': [],
    'Multinomial': [],
    'Neg': [],
    'NonMaxSuppression': [],
    'NonZero': [],
    'Not': [],
    'OneHot': [],
    'Or': [],
    'PRelu': [],
    'Pad': [],
    'Pow': [],
    'QLinearConv': [_string('auto_pad', 'NOTSET'), _ints('dilations', []), _int('group', 1), _ints('kernel_shape', []), _ints('pads', []), _ints('strides', [])],
    'QLinearMatMul': [],
    'QuantizeLinear': [_int('axis', 1)],
    'RNN': [_floats('activation_alpha',[]), _floats('activation_beta',[]), _strings('activations',['Tanh', 'Tanh']), _float('clip'), _string('direction', 'forward'), _int('hidden_size'), _int('layout',0)],
    'RandomNormal': [],
    'RandomNormalLike': [],
    'RandomUniform': [],
    'RandomUniformLike': [],
    'Reciprocal': [],
    'ReduceL1': [],
    'ReduceL2': [],
    'ReduceLogSum': [],
    'ReduceLogSumExp': [],
    'ReduceMax': [],
    'ReduceMean': [],
    'ReduceMin': [],
    'ReduceProd': [],
    'ReduceSum': [],
    'ReduceSumSquare': [],
    'Relu': [],
    'Reshape': [_int('allowzero', 0)],
    'Resize': [_string('coordinate_transformation_mode', 'half_pixel'),
               _float('cubic_coeff_a', -0.75),
               _int('exclude_outside', 0),
               _float('extrapolation_value', 0.0),
               _string('mode', 'nearest'),
               _string('nearest_mode', 'round_prefer_floor')],
    'ReverseSequence': [],
    'RoiAlign': [],
    'Round': [],
    'Scan': [],
    'Scatter (deprecated)': [],
    'ScatterElements': [],
    'ScatterND': [],
    'Selu': [_float('alpha', 1.67326), _float('gamma', 1.0507)],
    'SequenceAt': [],
    'SequenceConstruct': [],
    'SequenceEmpty': [],
    'SequenceErase': [],
    'SequenceInsert': [],
    'SequenceLength': [],
    'Shape': [_int('end'), _int('start', 0)],
    'Shrink': [],
    'Sigmoid': [],
    'Sign': [],
    'Sin': [],
    'Sinh': [],
    'Size': [],
    'Slice': [],
    'Softplus': [],
    'Softsign': [],
    'SpaceToDepth': [],
    'Split': [_int('axis', 0)],
    'SplitToSequence': [],
    'Sqrt': [],
    'Squeeze': [],
    'StringNormalizer': [],
    'Sub': [],
    'Sum': [],
    'Tan': [],
    'Tanh': [],
    'TfIdfVectorizer': [],
    'ThresholdedRelu': [],
    'Tile': [],
    'TopK': [],
    'Transpose': [_ints('perm', [])],
    'Trilu': [],
    'Unique': [],
    'Unsqueeze': [],
    'Upsample (deprecated)': [],
    'Where': [],
    'Xor': [],
    'Function': [],
    'DynamicQuantizeLinear': [],
    'GreaterOrEqual': [],
    'HardSwish': [],
    'LessOrEqual': [],
    'LogSoftmax': [],
    'MeanVarianceNormalization': [],
    'NegativeLogLikelihoodLoss': [],
    'Range': [],
    'Softmax': [],
    'SoftmaxCrossEntropyLoss': [],
    'QLinearAdd' : [],
    'QLinearGlobalAveragePool' : [_int('channels_Last')],
    'QLinearConcat' : [_int('axis')],
}



import sys
import numpy


def Cast(output_count, input, to):
    r"""
    Constrain input, output tensor type.
    Castable
        tensor(float16), tensor(float), tensor(double),
        tensor(int8), tensor(int16), tensor(int32), tensor(int64),
        tensor(uint8), tensor(uint16), tensor(uint32), tensor(uint64),
        tensor(bool), tensor(string), tensor(bfloat16)
    """

    cvtTable = {
        10: numpy.float16,
        1: numpy.float32,
        11: numpy.double,
        3: numpy.int8,
        5: numpy.int16,
        6: numpy.int32,
        7: numpy.int64,
        2: numpy.uint8,
        4: numpy.uint16,
        12: numpy.uint32,
        13: numpy.uint64,
        9: bool,
        8: str,
        16: object,
    }

    to_type = cvtTable[to]
    from_type = input.dtype

    if to_type is str:
        ss = []
        for i in input.flatten():
            s = str(i).encode('utf-8')
            su = s.decode('utf-8')
            ss.append(su)
        return numpy.array(ss).astype(numpy.object).reshape(input.shape)

    elif from_type is object or from_type is numpy.dtype(numpy.uint16) or to_type is object:
        little_endian = sys.byteorder == 'little'

        if object == to_type:
            np_uint16_view = input.astype(numpy.float32).flatten().view(dtype=numpy.uint16)
            np_bfp16 = np_uint16_view[1::2] if little_endian else np_uint16_view[0::2]
            return np_bfp16.reshape(input.shape).astype(numpy.uint16)
        else:
            np_bfp16 = input.flatten().view()
            np_fp32_zeros = numpy.zeros((len(np_bfp16) * 2,), dtype=numpy.uint16)
            if little_endian:
                np_fp32_zeros[1::2] = np_bfp16
            else:
                np_fp32_zeros[0::2] = np_bfp16

            np_fp32_from_bfloat = np_fp32_zeros.view(dtype=numpy.float32)
            return np_fp32_from_bfloat.reshape(input.shape)

    return input.astype(to_type)



import numpy as np
from .util import Iterator
from .util import _index_to_offset, _offset_to_index


# X: [DATA_BATCH, DATA_CHANNEL, DATA_FEATURE, DATA_FEATURE ...]
def MaxPool(output_count, X, auto_pad, ceil_mode, dilations, kernel_shape, pads, storage_order, strides):
    # feature dimension
    feature_dim = len(X.shape) - 2
    feature_shape = X.shape[2:]

    # default attribute setting
    if len(dilations) == 0:
        dilations = np.ones([feature_dim], dtype=np.int64)
    else:
        dilations = np.array(dilations)

    kernel_shape = np.array(kernel_shape)

    if len(pads) == 0:
        pads = np.zeros([feature_dim * 2], dtype=np.int64)
    else:
        pads = np.array(pads)

    if len(strides) == 0:
        strides = np.ones([feature_dim], dtype=np.int64)
    else:
        strides = np.array(strides)

    # output_spatial_shape
    output_shape = np.zeros([feature_dim], dtype=np.int64)

    if auto_pad == 'NOTSET':
        if ceil_mode == 0:  # ceil_mode
            for i in range(feature_dim):
                output_shape[i] = np.floor((X.shape[2 + i] + pads[i] + pads[i + feature_dim] -
                                           ((kernel_shape[i] - 1) * dilations[i] + 1)) / strides[i] + 1)
        else:
            for i in range(feature_dim):
                output_shape[i] = np.ceil((X.shape[2 + i] + pads[i] + pads[i + feature_dim] -
                                          ((kernel_shape[i] - 1) * dilations[i] + 1)) / strides[i] + 1)
    elif auto_pad == 'VALID':  # auto_pad(deprecated)
        for i in range(feature_dim):
            output_shape[i] = np.ceil((X.shape[2 + i] - ((kernel_shape[i] - 1) * dilations[i] + 1) + 1) / strides[i])
    elif auto_pad == 'SAME_LOWER' or auto_pad == 'SAME_UPPER':
        for i in range(feature_dim):
            output_shape[i] = np.ceil(X.shape[2 + i] / strides[i])
            pad = (output_shape[i] - 1) * strides[i] + ((kernel_shape[i] - 1) * dilations[i] + 1) - X.shape[2 + i]
            pads[i] = pads[i + feature_dim] = int(pad / 2)
            if pad % 2 == 1:
                if auto_pad == 'SAME_UPPER':
                    pads[i + feature_dim] += 1
                else:
                    pads[i] += 1

    # MaxPool
    Y = np.zeros([X.shape[0] * X.shape[1] * int(np.prod(output_shape))], dtype=X.dtype)
    Indices = np.zeros([X.shape[0] * X.shape[1] * int(np.prod(output_shape))], dtype=np.int64)
    y_idx = 0

    if storage_order == 1:
        if len(output_shape) >= 2:
            storage_unit = output_shape[-2] * output_shape[-1]
        else:
            storage_order = 0

    # Make dilation kernel
    new_kernel_shape = dilations * np.array(kernel_shape)
    if not np.any(dilations == 1):
        new_kernel_shape -= np.ones([feature_dim], dtype=np.int64)

    for batch in range(X.shape[0]):
        for channel in range(X.shape[1]):
            x_iter = Iterator(-pads[0:feature_dim], -pads[0:feature_dim] + output_shape * strides, strides)
            x = X[batch, channel]

            while x_iter.next():
                x_idx = x_iter.index

                # Make slicers for copy pactch of X[batch, channel] on to padded X
                x_padded_patch = np.zeros(new_kernel_shape)
                x_slicers, x_padded_slicers = [], []
                for i in range(feature_dim):
                    x_start = 0 if x_idx[i] < 0 else x_idx[i]
                    x_end = min(x.shape[i], x_idx[i] + new_kernel_shape[i])
                    x_slicers.append(slice(x_start, x_end, dilations[i]))

                    x_padded_start = -x_idx[i] if x_idx[i] < 0 else 0
                    x_padded_end = x_padded_start + (x_end - x_start)
                    x_padded_slicers.append(slice(x_padded_start, x_padded_end, dilations[i]))

                x_patch = x[tuple(x_slicers)]
                x_padded_patch[x_padded_slicers] = x[tuple(x_slicers)]
                k_idx = _offset_to_index(np.argmax(x_padded_patch), x_padded_patch.shape)

                y = np.max(x_patch)
                d_idx = x_idx + k_idx
                argmax_idx = _index_to_offset(d_idx, feature_shape)

                Y[y_idx] = y

                if storage_order == 1:
                    remainder = y_idx % storage_unit
                    share = y_idx - remainder
                    a = remainder * output_shape[-1]
                    Indices[share + a % storage_unit + int(a / storage_unit)] = argmax_idx
                else:
                    Indices[y_idx] = argmax_idx

                y_idx += 1

    y_shape = X.shape[0:2] + tuple(output_shape)

    Y = Y.reshape(y_shape)
    Indices = Indices.reshape(y_shape)

    return Y, Indices
